/**
 *  Copyright (c) 1997-2013, tinygroup.org (luo_guo@live.cn).
 *
 *  Licensed under the GPL, Version 3.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *       http://www.gnu.org/licenses/gpl.html
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 * --------------------------------------------------------------------------
 *  版权 (c) 1997-2013, tinygroup.org (luo_guo@live.cn).
 *
 *  本开源软件遵循 GPL 3.0 协议;
 *  如果您不遵循此协议，则不被允许使用此文件。
 *  你可以从下面的地址获取完整的协议文本
 *
 *       http://www.gnu.org/licenses/gpl.html
 */
package org.tinygroup.autolayout;

import java.util.ArrayList;

/**
 * 自动布局算法类
 * 
 * @author luoguo
 */
public class CustomDirectedGraphLayout {
	ArrayList<Node> nodeList;

	public CustomDirectedGraphLayout(ArrayList<Node> nodeList) {
		this.nodeList = nodeList;
	}

	/**
	 * 查找Start到END的非循环主线，然后排列其它的节点<br>
	 * 如果Start到END有多条主线，以路线长的为主
	 * 
	 * @throws Exception
	 */
	public ArrayList<Node> arrange() throws Exception {
		Node node = PointUtil.getStartNode(nodeList);
		if (node.linkList.size() == 0) {
			throw new Exception("Start节点没有后续节点");
		}
		node = PointUtil.getEndNode(nodeList);
		if (node.linkList.size() == 0) {
			throw new Exception("End节点没有前续节点");
		}
		ArrayList<ArrayList<Node>> allPath = findAllPath();
		System.out.printf("found process path's number:%d\n", allPath.size());
		ArrayList<ArrayList<Node>> arrangedList = new ArrayList<ArrayList<Node>>();
		ArrayList<Node> mainPath = arrangePath(allPath, arrangedList);

		// 以下用于对位置进行调整

		adjustNodePosition(mainPath, arrangedList);
		return mainPath;
	}

	/**
	 * 查找所有通路
	 * 
	 * @return
	 * @throws Exception
	 */
	private ArrayList<ArrayList<Node>> findAllPath() throws Exception {
		ArrayList<ArrayList<Node>> result = new ArrayList<ArrayList<Node>>();
		findPath(result, null, null);
		return result;
	}

	/**
	 * 查找所有的路径start---end的路径
	 * 
	 * @param result
	 * @param list
	 * @param node
	 * @throws Exception
	 */
	@SuppressWarnings("unchecked")
	private void findPath(ArrayList<ArrayList<Node>> result,
			ArrayList<Node> list, Node node) throws Exception {
		// 如果是开始，则赋值给Start
		if (node == null) {
			node = PointUtil.getStartNode(nodeList);
			list = new ArrayList<Node>();
		}
		// 如果已经在了，说明不是主线
		if (list.indexOf(node) < 0) {
			list.add(node);
		} else {
			return;
		}
		if (node == PointUtil.getEndNode(nodeList)) {
			result.add(list);
			return;
		}
		// 如果只有一个节点
		if (node.startLinkCount == 1) {
			findPath(result, list, node.getNextNode());
		} else {
			for (Link link : node.linkList) {
				if (link.start == node) {
					findPath(result, (ArrayList<Node>) list.clone(), link.end);
				}
			}
		}
	}

	/**
	 * 查找主线，即，从start 到end的最短路径(非循环)
	 * 
	 * @param allPath
	 *            所有的路径
	 * @param arrangedPath
	 *            已经排好序的路径
	 * @return 主线
	 * @throws Exception
	 */
	private ArrayList<Node> arrangePath(
			ArrayList<ArrayList<Node>> beArrangedPath,
			ArrayList<ArrayList<Node>> arrangedPath) throws Exception {

		// 计算并打印共享节点情况

		int[] shareNumbers = new int[beArrangedPath.size()];
		int maxShareNumbers = 0;
		for (int i = 0; i < beArrangedPath.size(); i++) {
			shareNumbers[i] = 0;
			for (int j = 0; j < beArrangedPath.size(); j++) {
				if (i != j) {
					ArrayList<Node> shareNodes = PointUtil.findSameNode(
							beArrangedPath.get(i), beArrangedPath.get(j));
					shareNumbers[i] += shareNodes.size();
					if (shareNumbers[i] > maxShareNumbers) {
						maxShareNumbers = shareNumbers[i];
					}
				}
			}
		}

		// 查找主线，主线就是与其他路径共享节点最多的路径，如果共享路径数相同，则路径中节点数最少的为主路径

		ArrayList<Node> path = null;

		for (int i = 0; i < shareNumbers.length; i++) {
			if (shareNumbers[i] == maxShareNumbers) {
				if (path == null) {
					path = beArrangedPath.get(i);
				} else {
					if (PointUtil.findSameNode(path, beArrangedPath.get(i))
							.size() <= 2) {// 如果两个没有相关，则长的为准
						if (beArrangedPath.get(i).size() > path.size()) {
							path = beArrangedPath.get(i);
						}
					} else {
						if (beArrangedPath.get(i).size() < path.size()) {
							path = beArrangedPath.get(i);
						}
					}
				}
			}
		}
		// 如果有另外一条线，全被此路径包含，此设置子线为主线
		for (ArrayList<Node> p : beArrangedPath) {
			boolean findAll = true;
			for (Node m : p) {
				if (path.indexOf(m) < 0) {
					findAll = false;
					break;
				}
			}
			if (findAll) {
				path = p;
			}
		}
		arrangedPath.add(path);
		beArrangedPath.remove(path);
		// 排定其它路径
		boolean putOnHead = false;
		while (beArrangedPath.size() > 0) {
			arrangeOtherPath(beArrangedPath, arrangedPath, putOnHead);
			putOnHead = !putOnHead;
		}

		return path;
	}

	/**
	 * 主线确定后的其它路径的安排策略，考虑与其它节点无共享节点的处理 算法思路：查找与arrangedPath关系最密切的路径，最放在它的旁边
	 */

	private void arrangeOtherPath(ArrayList<ArrayList<Node>> beArrangedPath,
			ArrayList<ArrayList<Node>> arrangedPath, boolean putOnHead)
			throws Exception {
		int shareNodesCount = 0;
		ArrayList<Node> path = null;
		// 找到与已经排列好的路径关联度最大的模块
		for (ArrayList<Node> bp : beArrangedPath) {
			int c = 0;
			for (ArrayList<Node> p : arrangedPath) {
				c += PointUtil.findSameNode(bp, p).size();
			}
			if (c > shareNodesCount) {
				shareNodesCount = c;
				path = bp;
			} else if (c == shareNodesCount) {// 如果两个共享模块数相同，则节点数多的优先
				if (bp.size() > path.size()) {
					path = bp;
				}
			}
		}
		if (putOnHead) {
			arrangedPath.add(0, path);
		} else {
			arrangedPath.add(path);
		}
		beArrangedPath.remove(path);
	}

	/**
	 * 调整位置
	 * 
	 * @param mainPath
	 * @param arrangedList
	 */
	private void adjustNodePosition(ArrayList<Node> mainPath,
			ArrayList<ArrayList<Node>> arrangedList) {
		// 确定主线以及非主线完整路径上点的坐标位置

		adjustMainAndOtherNodePosition(arrangedList, mainPath);
		// 然后以主线为基准进行调整

		// 不在主线上，但是在流程中的点的位置
		System.out.println("===========pos=================");
		// 首先取得还没有被定为的点
		ArrayList<Node> noMainPathPointNodeList = new ArrayList<Node>();
		for (Node m : nodeList) {
			if (m.point.x == 0)
				noMainPathPointNodeList.add(m);
		}
		int i = 0;
		// 如果还有节点没有被定为则继续定位
		// 限定定为次数为5次，防止出现死循环
		while (PointUtil.isConstainNoArangeNode(noMainPathPointNodeList)) {
			if (i == 5)
				break;
			arangeNoMainPathPoint(noMainPathPointNodeList);
			i++;
		}
		// @TODO 把能放水平的放水平
		// 调整主线和分支上能调整的位置
		adjustMainAndOtherPathAgain();
		// 如果水平线会经过某个点，则要降低一个
		adjustHorizontalPosition();

		// 如果垂直线会经过某个点，则要水平拉开
		adjustVerticalPosition(mainPath);

		// 如果END不在最低一层，则END下降
		// 如果Start不在最高一层，则除Start外，其他下降一层
		adjustStartAndEndPosition();

	}

	// 调整开始和结束位置
	private void adjustStartAndEndPosition() {
		Node end = null;
		Node start = null;
		int minY = Integer.MAX_VALUE;
		int maxY = Integer.MIN_VALUE;
		for (Node m : nodeList) {
			if (m.getType() == Node.TYPE_END) {
				end = m;
			} else if (m.point.y > maxY) {
				maxY = m.point.y;
			}
			if (m.getType() == Node.TYPE_START) {
				start = m;
			} else if (m.point.y < minY) {
				minY = m.point.y;
			}
		}
		if (end.point.y <= maxY) {
			end.point.y = maxY + 1;
		}
		if (start.point.y >= minY) {
			for (Node m : nodeList) {
				if (m != start) {
					m.point.y += minY - start.point.y + 1;
				}
			}
		}
	}

	/**
	 * 调整节点垂直方向位置
	 * 
	 * @param mainPath
	 */
	private void adjustVerticalPosition(ArrayList<Node> mainPath) {
		boolean moved = false;
		while (!moved) {
			moved = true;
			for (Node m : nodeList) {
				if (m.linkList.size() > 1) {// 如果一个节点的连接数大于1，且会导致重复，则下降与之水平相连节点

					int c = 0;
					int uc = 0, dc = 0;
					boolean hasLinkToUpNode = false;
					for (Link link : m.linkList) {
						Node om = link.getAnotherNode(m);
						if (om.point.x == m.point.x) {
							if (!hasLinkToUpNode) {// 如果已经发现，就不用再找了
								for (Link l : om.linkList) {
									if (l.getAnotherNode(om).point.y < m.point.y) {
										hasLinkToUpNode = true;
									}
								}
							}
							c++;
							if (om.point.y > m.point.y) {
								dc++;
							} else if (om.point.y < m.point.y) {
								uc++;
							}
						}
					}
					// 把它放在两个中间
					if (c >= 2) {
						if (uc == 2) {
							PointUtil.exchangePosition(m,
									PointUtil.getNearestUpNode(m));
							moved = false;
						} else if (dc == 2) {
							PointUtil.exchangePosition(m,
									PointUtil.getNearestDownNode(m));
							moved = false;
						} else if (uc > 2 || dc > 2) {
							if (m.point.getX() < mainPath.get(0).point.x) {
								PointUtil.moveLeft(m, nodeList);
								moved = false;
							} else if (m.point.getX() < mainPath.get(0).point.x) {
								PointUtil.moveRight(m, nodeList);
								moved = false;
							}
						}
					}
				}
			}
		}
	}

	/**
	 * 调整节点水平方向位置
	 */
	private void adjustHorizontalPosition() {
		boolean moved = false;
		while (!moved) {
			moved = true;
			for (Node m : nodeList) {
				if (m.linkList.size() > 1) {// 如果一个节点的连接数大于1，且会导致重复，则下降与之水平相连节点

					int c = 0;
					int lc = 0, rc = 0;
					boolean hasLinkToUpNode = false;
					for (Link link : m.linkList) {
						Node om = link.getAnotherNode(m);
						if (om.point.y == m.point.y) {
							if (!hasLinkToUpNode) {// 如果已经发现，就不用再找了
								for (Link l : om.linkList) {
									if (l.getAnotherNode(om).point.y < m.point.y) {
										hasLinkToUpNode = true;
									}
								}
							}
							c++;
							if (om.point.x > m.point.x) {
								rc++;
							} else if (om.point.x < m.point.x) {
								lc++;
							}
						}
					}
					if (c >= 2 && (lc >= 2 || rc >= 2)) {
						if (m.getType() == Node.TYPE_END || hasLinkToUpNode) {
							PointUtil.moveDown(m, nodeList);
						} else {
							for (Link link : m.linkList) {
								if (link.getAnotherNode(m).point.y == m.point.y) {
									PointUtil.moveDown(link.getAnotherNode(m),
											nodeList);
								}
							}
						}
						moved = false;
					}
				}
			}
		}
	}

	/**
	 * 主线和分支上横向和竖直位置调整
	 */
	private void adjustMainAndOtherPathAgain() {
		boolean moved = false;
		while (!moved) {
			moved = true;
			for (Node m : nodeList) {
				for (Link link : m.linkList) {
					Node n = link.getAnotherNode(m);
					if (n.getPoint().x != m.point.x) {// 如果m上的点与之连接的点的x值不同,end节点除外
						if (!PointUtil.isHasNodeInPosition(n.point.x,
								m.point.y, nodeList)) {
							n.point.setPosition(n.point.x, m.point.y);
						} else if (!PointUtil.isHasNodeInPosition(m.point.x,
								n.point.y, nodeList)) {
							m.point.setPosition(m.point.x, n.point.y);
						}
					}
					if (n.getPoint().y != m.point.y && m.type != Node.TYPE_END) {// 如果m上的点与之连接的点的y值不同
						if (!PointUtil.isHasNodeInPosition(n.point.x,
								m.point.y, nodeList)) {
							n.point.setPosition(n.point.x, m.point.y);
						} else if (!PointUtil.isHasNodeInPosition(m.point.x,
								n.point.y, nodeList)) {
							m.point.setPosition(m.point.x, n.point.y);
						}
					}
				}
			}
		}

		// 把应该对齐的对齐
		for (Node m : nodeList) {
			if (m.getMainNode() == Node.POSITION_SUB) {
				for (Link link : m.linkList) {
					Node d = link.getAnotherNode(m);
					if (d.getMainNode() == Node.POSITION_MAIN) {// 如果是连在主线上的，如果能调整则调整位置
						if (!PointUtil.isHasNodeInPosition(m.point.x,
								d.point.y, nodeList)) {
							m.point.setPosition(m.point.x, d.point.y);
						}
					}
				}
			}
		}

	}

	/**
	 * 确定主线以及其它start->end所以路径上节点位置
	 * 
	 * @param arrangedList
	 * @param mainPath
	 */
	private void adjustMainAndOtherNodePosition(
			ArrayList<ArrayList<Node>> arrangedList, ArrayList<Node> mainPath) {
		int mainPasPos = 0;
		for (int i = 0; i < arrangedList.size(); i++) {
			if (arrangedList.get(i) == mainPath) {
				mainPasPos = i;
			}
		}
		// 首先为主线赋初值
		for (int j = 0; j < mainPath.size(); j++) {
			Node m = mainPath.get(j);
			if (m.getPoint().getX() == 0) {// 如果没有赋过值，则赋值
				m.setMainNode(Node.POSITION_MAIN);
				m.getPoint().setPosition(mainPasPos + 1, j + 1);
			}
		}
		// 为所有在完整路径中的点赋值
		int x = mainPasPos + 1;
		for (int i = mainPasPos - 1; i >= 0; i--) {
			boolean find = false;
			for (int j = 0; j < arrangedList.get(i).size(); j++) {
				if (arrangedList.get(i) != mainPath) {
					Node m = arrangedList.get(i).get(j);
					if (m.getPoint().getX() == 0) {// 如果没有赋过值，则赋值
						m.getPoint().setPosition(i + 1, j + 1);
						m.setMainNode(Node.POSITION_SUB);
						find = true;
					}
				}
			}
			if (find) {
				x--;
			}
		}
		x = mainPasPos + 1;
		for (int i = mainPasPos + 1; i < arrangedList.size(); i++) {
			boolean find = false;
			for (int j = 0; j < arrangedList.get(i).size(); j++) {
				if (arrangedList.get(i) != mainPath) {
					Node m = arrangedList.get(i).get(j);
					if (m.getPoint().getX() == 0) {// 如果没有赋过值，则赋值
						m.getPoint().setPosition(x + 1, j + 1);
						m.setMainNode(Node.POSITION_SUB);
						find = true;
					}
				}
			}
			if (find) {
				x++;
			}
		}
		// 为所有在完整路径中的点赋值
		for (int i = 0; i < arrangedList.size(); i++) {
			for (int j = 0; j < arrangedList.get(i).size(); j++) {
				if (arrangedList.get(i) != mainPath) {
					Node m = arrangedList.get(i).get(j);
					if (m.getPoint().getX() == 0) {// 如果没有赋过值，则赋值
						m.getPoint().setPosition(i + 1, j + 1);
						m.setMainNode(Node.POSITION_SUB);
					}
				}
			}
		}

	}

	/**
	 * 为不在主线上路径上的点指定坐标
	 * 
	 * @param noMainPathPointNodeList
	 */
	private void arangeNoMainPathPoint(ArrayList<Node> noMainPathPointNodeList) {
		for (Node m : noMainPathPointNodeList) {
			if (m.point.y == 0) {
				if (m.linkList.size() == 0) {
					// @TODO 如果是离散的，则自动排列
				} else if (m.linkList.size() == 1
						&& m.linkList.get(0).getAnotherNode(m).point.y > 0) {// 如果只有一个连接，且连接已定位
					Node linkedNode = m.linkList.get(0).getAnotherNode(m);
					// 查找其周围位置，是否有空位，如果有则置之

					for (int i = -1; i <= 1; i += 2) {
						boolean isSet = false;
						for (int j = -1; j <= 1; j += 2) {
							if (!PointUtil.isHasNodeInPosition(
									linkedNode.point.x, linkedNode.point.y + j,
									nodeList)) {
								m.point.setPosition(linkedNode.point.x,
										linkedNode.point.y + j);
								isSet = true;
								break;
							} else if (!PointUtil.isHasNodeInPosition(
									linkedNode.point.x + i, linkedNode.point.y,
									nodeList)) {
								m.point.setPosition(linkedNode.point.x + i,
										linkedNode.point.y);
								isSet = true;
								break;
							} else if (!PointUtil.isHasNodeInPosition(
									linkedNode.point.x + i, linkedNode.point.y
											+ j, nodeList)) {
								m.point.setPosition(linkedNode.point.x + i,
										linkedNode.point.y + j);
								isSet = true;
								break;
							}
						}
						if (isSet)
							break;
					}
				} else {
					// 如果连接有多个连接，则查找已定位的最近的两个点，以此两点来定其位置
					int m_count = 0;// 与之连接且已定位模块数
					ArrayList<Node> m_list = new ArrayList<Node>();
					for (Link link : m.linkList) {
						if (link.getAnotherNode(m).point.y > 0) {
							m_count++;
							m_list.add(link.getAnotherNode(m));
						}
					}
					if (m_count == 1) {
						// --------modify by
						// luoguo----2009-8-11----------------------------
						// 这里应该从m_list中取，而不是linkList
						// Node linkedNode =
						// m.linkList.get(0).getAnotherNode(m);
						Node linkedNode = m_list.get(0);
						// 查找其周围位置，是否有空位，如果有则置之
						for (int i = -1; i <= 1; i += 2) {
							for (int j = -1; j <= 1; j += 2) {
								if (!PointUtil.isHasNodeInPosition(
										linkedNode.point.x + i,
										linkedNode.point.y + j, nodeList)) {
									m.point.setPosition(linkedNode.point.x + i,
											linkedNode.point.y + j);
								}
							}
						}
					} else {// 如果有多个，则查找最近的两个
						for (int pos1 = 0; pos1 < m_list.size(); pos1++) {
							for (int pos2 = 0; pos2 < m_list.size(); pos2++) {
								if (pos1 != pos2) {
									// 此时应该在pos1及pos2里保存着距离最近的两个模块
									if (!PointUtil.isHasNodeInPosition(
											m_list.get(pos1).point.x,
											m_list.get(pos2).point.y, nodeList)) {
										m.point.setPosition(
												m_list.get(pos1).point.x,
												m_list.get(pos2).point.y);
										break;
									} else if (!PointUtil.isHasNodeInPosition(
											m_list.get(pos2).point.x,
											m_list.get(pos1).point.y, nodeList)) {
										m.point.setPosition(
												m_list.get(pos2).point.x,
												m_list.get(pos1).point.y);
										break;
									} else {
										// @TODO 如果最应该放置的两个点都有，如何放置
										//
										// System.out.println("");
									}
								}
							}
							if (m.point.y > 0) {// 如果已经设置过，则跳出
								break;
							}
						}
					}
				}
			}
		}
	}
}
